/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SoapComponent.java 154 25 sept. 06 alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap;

import java.io.File;
import java.util.Iterator;
import java.util.logging.Level;

import javax.jbi.JBIException;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.ConfigurationContextFactory;
import org.apache.axis2.description.AxisModule;
import org.apache.axis2.description.Flow;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.phaseresolver.PhaseException;
import org.apache.axis2.phaseresolver.PhaseHolder;
import org.apache.axis2.phaseresolver.PhaseMetadata;
import org.objectweb.petals.binding.soap.listener.incoming.SoapExternalListenerManager;
import org.objectweb.petals.component.framework.bc.AbstractBindingComponent;
import org.objectweb.petals.component.framework.su.ServiceUnitManager;

/**
 * The SOAP binding component.
 *
 * @author alouis - EBM Websourcing
 * @author chamerling - EBM Websourcing
 *
 */
public class SoapComponent extends AbstractBindingComponent {

    /**
     * The listener processing SOAP request providing from external clients.
     * Incoming SOAP requests.
     */
    protected SoapExternalListenerManager externalListenerManager;

    /**
     * The SOAP component context
     */
    protected SoapComponentContext soapContext;

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.AbstractComponent#doStart()
     */
    @Override
    public void doStart() throws JBIException {
        try {
            externalListenerManager.start();
        } catch (AxisFault e) {
            getLogger().severe(e.getMessage());
            throw new JBIException(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.AbstractComponent#doStop()
     */
    @Override
    public void doStop() throws JBIException {
        try {
            externalListenerManager.stop();
        } catch (AxisFault e) {
            getLogger().severe(e.getMessage());
            throw new JBIException(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.AbstractComponent#doInit()
     */
    @Override
    protected void doInit() throws JBIException {

        // create the SOAP component context
        this.soapContext = new SoapComponentContext(getLogger(), getComponentConfiguration());

        // the axis2.xml is required
        File axis2File = new File(getContext().getInstallRoot(), "axis2.xml");
        if (!axis2File.exists()) {
            throw new JBIException("Can not get axis2.xml configuration file");
        }

        // get axis configuration context
        this.getLogger().fine("Creating Axis configuration context...");
        ConfigurationContext axisConfigurationContext = null;
        try {
            axisConfigurationContext = ConfigurationContextFactory
                    .createConfigurationContextFromFileSystem(getContext()
                            .getInstallRoot(), axis2File.getAbsolutePath());
            
            this.getLogger().fine("Axis configuration context created.");
        } catch (AxisFault e) {
            getLogger().log(Level.WARNING, e.getMessage());
            throw new JBIException("Can not initialize SOAP BC", e);
        }

        // set SOAP Server
        this.externalListenerManager = new SoapExternalListenerManager(this,
                getChannel(), (ServiceUnitManager) getServiceUnitManager(),
                axisConfigurationContext, soapContext,
                getComponentConfiguration(), getLogger());

        // for each module, we add its flows (if any) to the corresponding axis
        // phases
        this.engageModulesHandlers(axisConfigurationContext.getAxisConfiguration());

        // set service unit listener
        SoapSUListener suListener = new SoapSUListener(soapContext,
                axisConfigurationContext, getContext(), externalListenerManager
                .getServicesURL(), getLogger());
        this.setServiceUnitListener(suListener);
    }

    /**
     * @param axisConfiguration
     * @throws JBIException
     */
    @SuppressWarnings("unchecked")
    private void engageModulesHandlers(AxisConfiguration axisConfiguration)
            throws JBIException {
        Iterator<String> moduleNames = axisConfiguration.getModules().keySet()
                .iterator();
        while (moduleNames.hasNext()) {
            try {
                this.engageModuleHandlersIntoPhases(axisConfiguration
                        .getModule(moduleNames.next()), axisConfiguration);
            } catch (PhaseException e) {
                this.getLogger().log(Level.WARNING, e.getMessage());
                throw new JBIException("Can not initialize SOAP BC", e);
            }
        }
        this.getLogger().fine("Axis modules flows added");
    }

    /**
     * For each Axis phase, we add the corresponding module's handlers if any.
     *
     * @param module
     *            the module to process
     * @param axisConfiguration
     *            the axis configuration
     * @throws PhaseException
     *             can the thrown by method addHandler()
     */
    private void engageModuleHandlersIntoPhases(AxisModule module,
            AxisConfiguration axisConfiguration) throws PhaseException {
        Flow moduleFlow = null;
        PhaseHolder phaseHolder = null;

        this.getLogger().fine("Engaging module " + module.getName() + " in flows");

        // for each phase, we add the corresponding module's handlers if any
        for (int type = PhaseMetadata.IN_FLOW; type <= PhaseMetadata.FAULT_OUT_FLOW; type++) {

            if (type == PhaseMetadata.IN_FLOW) {

                phaseHolder = new PhaseHolder(axisConfiguration
                        .getInFlowPhases());
                moduleFlow = module.getInFlow();

            } else if (type == PhaseMetadata.OUT_FLOW) {

                phaseHolder = new PhaseHolder(axisConfiguration
                        .getOutFlowPhases());
                moduleFlow = module.getOutFlow();

            } else if (type == PhaseMetadata.FAULT_IN_FLOW) {

                phaseHolder = new PhaseHolder(axisConfiguration
                        .getInFaultFlowPhases());
                moduleFlow = module.getFaultInFlow();

            } else if (type == PhaseMetadata.FAULT_OUT_FLOW) {

                phaseHolder = new PhaseHolder(axisConfiguration
                        .getOutFaultFlowPhases());
                moduleFlow = module.getFaultOutFlow();
            }

            if (moduleFlow != null && phaseHolder != null) {
                for (int j = 0; j < moduleFlow.getHandlerCount(); j++) {
                    phaseHolder.addHandler(moduleFlow.getHandler(j));
                }
            }
        }
    }

    /**
     *
     * @return
     */
    public SoapComponentContext getSoapContext() {
        return soapContext;
    }

    /**
     *
     * @return
     */
    public SoapExternalListenerManager getExternalListenerManager() {
        return externalListenerManager;
    }
}
