/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.deployment.DeploymentConstants;
import org.apache.axis2.deployment.DeploymentErrorMsgs;
import org.apache.axis2.deployment.DeploymentException;
import org.apache.axis2.deployment.ServiceBuilder;
import org.apache.axis2.description.AxisModule;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.i18n.Messages;
import org.apache.axis2.util.XMLUtils;
import org.objectweb.petals.binding.soap.listener.SUProperties;
import org.objectweb.petals.binding.soap.listener.incoming.PetalsAxisService;
import org.objectweb.petals.binding.soap.util.SoapSUClassLoader;
import org.objectweb.petals.component.framework.ExtensionsConstants;
import org.objectweb.petals.component.framework.PEtALSCFException;
import org.objectweb.petals.component.framework.su.ServiceUnitListener;
import org.objectweb.petals.component.framework.util.ClassLoaderUtil;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.PropertiesUtil;
import org.objectweb.petals.jbi.descriptor.Consumes;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.jbi.descriptor.Provides;
import org.objectweb.petals.jbi.descriptor.ServiceUnitExtensibleElement;
import org.objectweb.petals.ws.fault.WsnFault;
import org.objectweb.petals.ws.notification.WsnManager;

/**
 * A service unit listener used to register new service into Axis Engine during
 * SU deployment.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class SoapSUListener implements ServiceUnitListener {

    /**
     * The axis engine configuration context
     */
    private ConfigurationContext axisConfigurationContext;

    /**
     * The JBI component context
     */
    private ComponentContext componentContext;

    /**
     * The SOAP component context (configuration values)
     */
    private SoapComponentContext soapContext;

    /**
     * JDK logger
     */
    private Logger logger;

    /**
     * The SU root path
     */
    private String suRootPath;

    /**
     *
     */
    private String servicesPrefix;

    /**
     * Creates a new instance of {@link SoapSUListener}
     *
     * @param soapContext
     * @param axisConfigurationContext
     * @param componentContext
     * @param servicesURL
     * @param logger
     */
    public SoapSUListener(SoapComponentContext soapContext,
            ConfigurationContext axisConfigurationContext,
            ComponentContext componentContext, String servicesURL, Logger logger) {
        this.soapContext = soapContext;
        this.axisConfigurationContext = axisConfigurationContext;
        this.componentContext = componentContext;
        this.servicesPrefix = servicesURL;
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUDeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUDeployed(final String serviceUnitName,
            final String suRootPath, final JBIDescriptor descriptor)
            throws PEtALSCFException {
        this.logger.log(Level.FINE, "Deploying a new AxisService for SU "
                + serviceUnitName);

        // we set the su rootpath
        this.suRootPath = suRootPath;

        this.soapContext.addJbiDescriptor(serviceUnitName, descriptor);

        // check if a services.xml file is available
        File servicesXml = new File(suRootPath, "services.xml");
        if (servicesXml.exists()) {
            this.soapContext.addServiceDescriptor(serviceUnitName, servicesXml);
        }

        // add the modules references to the context
        this.addModulesToContext(descriptor);

        // add the parameters for the associated service to the context
        this.setServiceParametersToContext(descriptor);

        /*
         * Create a new Topic on each provide node if the address starts with
         * the petals topic prefix.
         */
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            this.createTopic(provides);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUStarted(java.lang.String)
     */
    public void onSUStarted(String serviceUnitName) throws PEtALSCFException {

        // get descriptors
        JBIDescriptor descriptor = this.soapContext
                .getJbiDescriptor(serviceUnitName);
        File serviceDescription = this.soapContext
                .getServiceDescriptor(serviceUnitName);

        if (descriptor != null) {

            /*
             * Register new Axis2 service only on consumes nodes. Create an
             * AxisService for each requested endpoint. If an enpoint has not
             * been found, simply log a severe error. TODO : If a service
             * creation fails, and if services have been already registered for
             * this service unit, unregister them.
             */
            List<Consumes> consumesList = descriptor.getServices()
                    .getConsumes();
            for (Consumes consumes : consumesList) {
                this.createAxisService(consumes, serviceDescription);
            }

            List<Provides> providesList = descriptor.getServices()
                    .getProvides();
            for (Provides provides : providesList) {
                this.createWSAccess(provides);
            }
        }
    }

    /**
     * Create an external web service provider
     *
     * @param provides
     */
    protected void createWSAccess(Provides provides) {
        Extensions cdkExtensions = getExtensions(provides);

        if (cdkExtensions != null) {
            Properties props = cdkExtensions.getPetalsExtensions();
            String externalWS = props.getProperty(SUProperties.ADDRESS);

            logger.info("Providing access to external web service '"
                    + externalWS + "'");
        }

    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUStopped(java.lang.String)
     */
    public void onSUStopped(final String serviceUnitName)
            throws PEtALSCFException {

        JBIDescriptor descriptor = this.soapContext
                .getJbiDescriptor(serviceUnitName);
        if (descriptor != null) {
            // delete registered axis services
            List<Consumes> consumesList = descriptor.getServices()
                    .getConsumes();
            for (Consumes consumes : consumesList) {
                this.unregisterAxisService(consumes);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUUndeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUUndeployed(String serviceUnitName, String suRootPath,
            JBIDescriptor descriptor) throws PEtALSCFException {

        soapContext.removeJbiDescriptor(serviceUnitName);
        soapContext.removeServiceDescriptor(serviceUnitName);

        removeModulesFromContext(descriptor);

        removeServiceParametersFromContext(descriptor);

        /*
         * Create a new Topic on each provide node if the address starts with
         * the petals topic prefix.
         */
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            this.deleteTopic(provides);
        }
    }

    /**
     * Create the Axis service that will handle incoming SOAP calls.
     *
     * @param consumes
     * @param serviceDescriptor
     * @throws PEtALSComponentSDKException
     */
    private void createAxisService(Consumes consumes, File serviceDescriptor)
            throws PEtALSCFException {
        AxisService axisService = null;

        // Get the created service name
        Extensions cdkExtensions = getExtensions(consumes);
        Properties props = cdkExtensions.getPetalsExtensions();
        String newServiceName = props.getProperty(SUProperties.ADDRESS);

        if (newServiceName == null) {
            throw new PEtALSCFException(
                    "Can not create an Axis service with a null name, the SU should provide an address extension value");
        }

        AxisConfiguration axisConfig = axisConfigurationContext
                .getAxisConfiguration();

        try {
            // try to find if the axisService has already been registered
            axisService = axisConfig.getService(newServiceName);

            if (axisService == null) {
                axisService = new PetalsAxisService(newServiceName,
                        componentContext, consumes, getExtensions(consumes), servicesPrefix, logger);

                // we set the associated service parameters
                String serviceParameters = this.soapContext
                        .getServiceParameters(newServiceName);
                if (serviceParameters != null) {
                    try {
                        OMElement parametersElements = buildParametersOM(serviceParameters);

                        // get an iterator on all <parameter> children
                        Iterator<OMElement> itr = parametersElements
                                .getChildrenWithName(new QName(
                                        DeploymentConstants.TAG_PARAMETER));

                        // iterate on parameters and set them to the associated
                        // axisService
                        while (itr.hasNext()) {
                            OMElement parameterElement = itr.next();

                            if (DeploymentConstants.TAG_PARAMETER
                                    .equalsIgnoreCase(parameterElement
                                            .getLocalName())) {
                                axisService
                                        .addParameter(getParameter(parameterElement));
                            }
                        }
                    } catch (XMLStreamException e) {
                        logger.warning(e.getMessage());
                    } catch (DeploymentException e) {
                        logger.warning(e.getMessage());
                    }
                }

                // We prepare a list for the modules ClassLoaders
                List<ClassLoader> suModulesClassLoaders = null;

                /*
                 * Add the modules for this new service. If the services.xml
                 * file if available, the modules references can be redefined
                 */
                List<String> modules = soapContext.getModules(newServiceName);
                if (modules != null) {
                    suModulesClassLoaders = new ArrayList<ClassLoader>();
                    for (String module : modules) {
                        AxisModule axisModule = axisConfig.getModule(module);
                        if (axisModule == null) {
                            throw new DeploymentException(Messages.getMessage(
                                    DeploymentErrorMsgs.MODULE_NOT_FOUND,
                                    module));
                        }
                        axisService.addModuleref(axisModule.getName());

                        // We add the current module ClassLoader to the list
                        suModulesClassLoaders.add(axisModule
                                .getModuleClassLoader());
                    }
                }

                // We set the service ClassLoader = initial ClassLoader
                // + jars of suRootPath + modules ClassLoaders
                axisService.setClassLoader(new SoapSUClassLoader(
                        ClassLoaderUtil.getUrls(suRootPath),
                        suModulesClassLoaders, axisService.getClassLoader()));

                // Try to load axis configuration from services.xml
                if (serviceDescriptor != null) {
                    try {
                        FileInputStream fis = new FileInputStream(
                                serviceDescriptor);
                        ServiceBuilder builder = new ServiceBuilder(fis,
                                axisConfigurationContext, axisService);
                        builder.populateService(builder.buildOM());
                    } catch (XMLStreamException e) {
                        logger.warning(e.getMessage());
                    } catch (FileNotFoundException e) {
                        logger.warning(e.getMessage());
                    } catch (DeploymentException e) {
                        logger.warning(e.getMessage());
                    }
                }

                // populate service with service descriptor
                QName jbiServiceQName = consumes.getServiceName();
                if (jbiServiceQName != null) {
                    axisService.setTargetNamespace(jbiServiceQName
                            .getNamespaceURI());
                }
                axisService.addParameter(new Parameter(Constants.SERVICE_CLASS,
                        "PetalsReceiver"));
                axisService
                        .addParameter(new Parameter(
                                Constants.Configuration.SEND_STACKTRACE_DETAILS_WITH_FAULTS,
                                Boolean.TRUE));

                axisConfig.addService(axisService);
            } else {
                logger.log(
                        Level.WARNING,
                        "The service '" + axisService.getName()
                        + "' is already registered in Axis, you can not register it twice");
            }
        } catch (AxisFault e) {
            throw new PEtALSCFException(
                    "Can not register Service into Axis context", e);
        }

        logger.log(Level.INFO, "The Axis service '" + newServiceName
                + "' has been registered and is available at '"
                + servicesPrefix + "/" + newServiceName + "'");
    }

    /**
     * Process the parameterElement object from the OM, and returns the
     * corresponding Parameter.
     *
     * @param parameterElement
     *            <code>OMElement</code>
     * @return the Parameter parsed
     * @throws DeploymentException
     *             if bad paramName
     */
    private Parameter getParameter(OMElement parameterElement)
            throws DeploymentException {

        Parameter parameter = new Parameter();
        // setting parameterElement
        parameter.setParameterElement(parameterElement);
        // setting parameter Name
        OMAttribute paramName = parameterElement.getAttribute(new QName(
                DeploymentConstants.ATTRIBUTE_NAME));
        if (paramName == null) {
            throw new DeploymentException(Messages.getMessage(
                    DeploymentErrorMsgs.BAD_PARAMETER_ARGUMENT,
                    parameterElement.toString()));
        }
        parameter.setName(paramName.getAttributeValue());
        // setting parameter Value (the child element of the parameter)
        OMElement paramValue = parameterElement.getFirstElement();
        if (paramValue != null) {
            parameter.setValue(paramValue);
            parameter.setParameterType(Parameter.OM_PARAMETER);
        } else {
            String paratextValue = parameterElement.getText();

            parameter.setValue(paratextValue);
            parameter.setParameterType(Parameter.TEXT_PARAMETER);
        }

        return parameter;
    }

    /**
     * Creates the OMElement corresponding to the parameters String, included in
     * parameters tags.
     *
     * @param parameters
     *            the parameters
     * @return Returns <code>OMElement</code> .
     * @throws javax.xml.stream.XMLStreamException
     *
     */
    private final OMElement buildParametersOM(String parameters)
            throws XMLStreamException {
        if (parameters != null) {
            parameters = "<parameters>" + parameters + "</parameters>";
        }
        OMElement element = (OMElement) XMLUtils.toOM(new StringReader(
                parameters));
        element.build();
        return element;
    }

    /**
     * Unregister the service from Axis.
     *
     * @param serviceQName
     * @throws PEtALSComponentSDKException
     */
    private void unregisterAxisService(Consumes consumes) 
            throws PEtALSCFException {

        Extensions cdkExtensions = getExtensions(consumes);
        Properties props = cdkExtensions.getPetalsExtensions();
        String endPointName = props.getProperty(SUProperties.ADDRESS);

        AxisConfiguration axisConfig = axisConfigurationContext
                      .getAxisConfiguration();

        logger.log(Level.INFO, "Removing Axis service '" + endPointName + "'");

        try {
            // register an axis service to axis engine
            AxisService axisService = axisConfig.getService(endPointName);

            // Feature request : #306664
            // FIXME : We do not have to remove the service group. It is
            // temporary until the next version of Axis2 where the removeService
            // method will fix it.
            if (axisService != null) {
                axisConfig.removeServiceGroup(endPointName);
                axisService.getAxisConfiguration().removeService(endPointName);
            } else {
                logger.log(Level.WARNING, "Service '" + endPointName
                        + "' not found, can not be unregistered from Axis2");
            }

        } catch (AxisFault e) {
            throw new PEtALSCFException(
                    "Can not remove service from Axis context", e);
        }
    }

    /**
     * Add the modules to the soap context.
     *
     * @param serviceUnitName
     * @param descriptor
     */
    private void addModulesToContext(JBIDescriptor descriptor) {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            this.addModules(getExtensions(provides));
        }

        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            this.addModules(getExtensions(consumes));
        }
    }

    /**
     * Add the parameters to the associated service.
     *
     * @param serviceUnitName
     * @param descriptor
     *            JBIDescriptor
     */
    private void setServiceParametersToContext(JBIDescriptor descriptor) {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            this.setServiceParameters(getExtensions(provides));
        }

        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            this.setServiceParameters(getExtensions(consumes));
        }
    }

    /**
     * Remove the modules from the soap context.
     *
     * @param descriptor
     *            JBIDescriptor
     */
    private void removeModulesFromContext(JBIDescriptor descriptor) {
        this.removeModulesOrServiceParametersFromContext(descriptor, true);
    }

    /**
     * Remove the modules from the soap context.
     *
     * @param descriptor
     *            JBIDescriptor
     */
    private void removeServiceParametersFromContext(JBIDescriptor descriptor) {
        this.removeModulesOrServiceParametersFromContext(descriptor, false);
    }

    /**
     * Remove the modules or the service parameters from the soap context.
     *
     * @param descriptor
     *            JBIDescriptor
     * @param removeModules
     *            remove the modules if set to true, otherwise remove the
     *            serviceParameters
     */
    private void removeModulesOrServiceParametersFromContext(
            JBIDescriptor descriptor, boolean removeModules) {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            if (removeModules) {
                this.removeModules(getExtensions(provides));
            } else {
                this.removeServiceParameters(getExtensions(provides));
            }
        }
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            if (removeModules) {
                this.removeModules(getExtensions(consumes));
            } else {
                this.removeServiceParameters(getExtensions(consumes));
            }
        }
    }

    /**
     * Add modules to the context.
     *
     * @param provides
     */
    private void addModules(Extensions extensions) {
        Properties props = extensions.getPetalsExtensions();

        String address = props.getProperty(SUProperties.ADDRESS);

        // get modules from extension
        String token = props.getProperty("modules");

        if (token != null) {
            // get individual modules values
            StringTokenizer st = new StringTokenizer(token, ",");
            List<String> mods = new ArrayList<String>(st.countTokens() + 1);
            mods.add(Constants.MODULE_ADDRESSING);
            while (st.hasMoreTokens()) {
                mods.add(st.nextToken().trim());
            }

            this.soapContext.addModules(address, mods);
        }
    }

    /**
     * Add parameters for the associated service to the componentContext.
     *
     * @param extensions
     */
    private void setServiceParameters(Extensions extensions) {
        Properties props = extensions.getPetalsExtensions();

        String address = props.getProperty(SUProperties.ADDRESS);

        // get serviceParameters from extension
        String serviceParameters = props.getProperty("service-parameters");

        if (address != null && serviceParameters != null) {

            this.soapContext.setServicePamameters(address, serviceParameters);
        }
    }

    /**
     * Remove the modules from the context.
     *
     * @param extensions
     */
    private void removeModules(Extensions extensions) {
        this.removeModulesOrServiceParameters(extensions, true);
    }

    /**
     * Remove the modules from the context.
     *
     * @param extensions
     */
    private void removeServiceParameters(Extensions extensions) {
        this.removeModulesOrServiceParameters(extensions, false);
    }

    /**
     * Remove the modules or the service parameters from the context.
     *
     * @param extensions
     * @param removeModules
     *            remove the modules if set to true, otherwise remove the
     *            serviceParameters
     */
    private void removeModulesOrServiceParameters(Extensions extensions,
            boolean removeModules) {
        Properties props = extensions.getPetalsExtensions();
        String address = props.getProperty(SUProperties.ADDRESS);

        if (removeModules) {
            this.soapContext.removeModules(address);
        } else {
            this.soapContext.removeServiceParameters(address);
        }
    }

    /**
     * Create a new topic on SU deployment
     *
     * @param topicName
     */
    protected void createTopic(Provides provides) {
        Extensions cdkExtensions = getExtensions(provides);
        Properties props = cdkExtensions.getPetalsExtensions();
        String address = props.getProperty(SUProperties.ADDRESS);

        WsnManager notificationManager = soapContext.getWsnManager();

        if (address.startsWith(SUProperties.TOPIC_PREFIX + ":")) {
            String topicName = SUProperties
                    .getTopicNameFromExtensionAddress(address);

            logger.info("Creating new topic '" + topicName + "'");
            try {
                notificationManager.addTopic(new QName(topicName));
            } catch (WsnFault e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }

    /**
     * Delete the topic on SU undeployment
     *
     * @param topicName
     */
    protected void deleteTopic(Provides provides) {
        // TopicSet topics = soapContext.getTopics();
        //
        // String address = PetalsExtensionsUtil
        // .extractValueFromKeyValueExtension(provides.getExtensions(),
        // SUProperties.ADDRESS);
        //
        // if (address.startsWith(Topic.TOPIC_PREFIX + ":")) {
        // String topicName = SUProperties
        // .getTopicNameFromExtensionAddress(address);
        // logger.info("Deleting topic '" + topicName + "'");
        // // TODO : Notify topic consumers ?
        // topics.removeTopic(topicName);
        // }
    }

    /**
     * This util method allows to extract extensions from a Provides or Consumes
     * SU sub element. If extensions contains place holders (references to
     * external file properties), they are automaticaly replaced by real value
     *
     * @param extensibleElement
     */
    public Extensions getExtensions(ServiceUnitExtensibleElement extensibleElement) {
        Extensions extensions = new Extensions(extensibleElement.getExtensions());
        String propertiesFilePath = soapContext.getComponentConfiguration().getProperties()
                .getProperty(ExtensionsConstants.PROPERTIES_FILE);
        if (propertiesFilePath != null && !"".equals(propertiesFilePath)) {
            try {
                extensions.setPetalsExtensions(PropertiesUtil.fillPlaceHolder(propertiesFilePath,
                        extensions.getPetalsExtensions()));
            } catch (PEtALSCFException e) {
                logger.log(Level.WARNING, "Cannot fill place holder of SU extensions", e);
            }
        }
        return extensions;
    }
}
